@@ -3,23 +3,31 @@ Implementation of the nested logit model, see page 86 of the book
 "discrete choice methods with simulation" by Train. for more details.
 
 Author: Tianyu Du
-Update; Apr. 28, 2022
+Update; Apr. 7, 2023
 """
 from typing import Dict, List, Optional
 
 import torch
 import torch.nn as nn
-from torch_choice.model.coefficient import Coefficient
+
 from torch_choice.data.choice_dataset import ChoiceDataset
+from torch_choice.data.joint_dataset import JointDataset
+from torch_choice.model.coefficient import Coefficient
+from torch_choice.model.formula_parser import parse_formula
 
 
 class NestedLogitModel(nn.Module):
     def __init__(self,
-                 category_to_item: Dict[object, List[int]],
-                 category_coef_variation_dict: Dict[str, str],
-                 category_num_param_dict: Dict[str, int],
-                 item_coef_variation_dict: Dict[str, str],
-                 item_num_param_dict: Dict[str, int],
+                 nest_to_item: Dict[object, List[int]],
+                 # method 1: specify variation and num param. dictionary.
+                 nest_coef_variation_dict: Optional[Dict[str, str]]=None,
+                 nest_num_param_dict: Optional[Dict[str, int]]=None,
+                 item_coef_variation_dict: Optional[Dict[str, str]]=None,
+                 item_num_param_dict: Optional[Dict[str, int]]=None,
+                 # method 2: specify formula and dataset.
+                 item_formula: Optional[str]=None,
+                 nest_formula: Optional[str]=None,
+                 dataset: Optional[JointDataset]=None,
                  num_users: Optional[int]=None,
                  shared_lambda: bool=False,
                  regularization: Optional[str]=None,
@@ -28,31 +36,42 @@ class NestedLogitModel(nn.Module):
         """Initialization method of the nested logit model.
 
         Args:
-            category_to_item (Dict[object, List[int]]): a dictionary maps a category ID to a list
-                of items IDs of the queried category.
+            nest_to_item (Dict[object, List[int]]): a dictionary maps a nest ID to a list
+                of items IDs of the queried nest.
 
-            category_coef_variation_dict (Dict[str, str]): a dictionary maps a variable type
+            nest_coef_variation_dict (Dict[str, str]): a dictionary maps a variable type
                 (i.e., variable group) to the level of variation for the coefficient of this type
                 of variables.
-            category_num_param_dict (Dict[str, int]): a dictionary maps a variable type name to
+            nest_num_param_dict (Dict[str, int]): a dictionary maps a variable type name to
                 the number of parameters in this variable group.
 
-            item_coef_variation_dict (Dict[str, str]): the same as category_coef_variation_dict but
+            item_coef_variation_dict (Dict[str, str]): the same as nest_coef_variation_dict but
                 for item features.
-            item_num_param_dict (Dict[str, int]): the same as category_num_param_dict but for item
+            item_num_param_dict (Dict[str, int]): the same as nest_num_param_dict but for item
                 features.
 
+            {nest, item}_formula (str): a string representing the utility formula for the {nest, item} level logit model.
+                The formula consists of '(variable_name|variation)'s separated by '+', for example:
+                "(var1|item) + (var2|user) + (var3|constant)"
+                where the first part of each term is the name of the variable
+                and the second part is the variation of the coefficient.
+                The variation can be one of the following:
+                'constant', 'item', 'item-full', 'user', 'user-item', 'user-item-full'.
+                All spaces in the formula will be ignored, hence please do not use spaces in variable/observable names.
+            dataset (JointDataset): a JointDataset object for training the model, the parser will infer dimensions of variables
+                and sizes of coefficients for the nest level model from dataset.datasets['nest']. The parser will infer dimensions of variables and sizes of coefficients for the item level model from dataset.datasets['item'].
+
             num_users (Optional[int], optional): number of users to be modelled, this is only
                 required if any of variable type requires user-specific variations.
                 Defaults to None.
 
             shared_lambda (bool): a boolean indicating whether to enforce the elasticity lambda, which
-                is the coefficient for inclusive values, to be constant for all categories.
-                The lambda enters the category-level selection as the following
-                Utility of choosing category k = lambda * inclusive value of category k
-                                               + linear combination of some other category level features
-                If set to True, a single lambda will be learned for all categories, otherwise, the
-                model learns an individual lambda for each category.
+                is the coefficient for inclusive values, to be constant for all nests.
+                The lambda enters the nest-level selection as the following
+                Utility of choosing nest k = lambda * inclusive value of nest k
+                                               + linear combination of some other nest level features
+                If set to True, a single lambda will be learned for all nests, otherwise, the
+                model learns an individual lambda for each nest.
                 Defaults to False.
 
             regularization (Optional[str]): this argument takes values from {'L1', 'L2', None}, which specifies the type of
@@ -67,22 +86,46 @@ class NestedLogitModel(nn.Module):
                 is not None.
                 Defaults to None.
         """
+        # handle nest level model.
+        using_formula_to_initiate = (item_formula is not None) and (nest_formula is not None)
+        if using_formula_to_initiate:
+            # make sure that the research does not specify duplicated information, which might cause conflict.
+            if (nest_coef_variation_dict is not None) or (item_coef_variation_dict is not None):
+                raise ValueError('You specify the {item, nest}_formula to initiate the model, you should not specify the {item, nest}_coef_variation_dict at the same time.')
+            if (nest_num_param_dict is not None) or (item_num_param_dict is not None):
+                raise ValueError('You specify the {item, nest}_formula to initiate the model, you should not specify the {item, nest}_num_param_dict at the same time.')
+            if dataset is None:
+                raise ValueError('Dataset is required if {item, nest}_formula is specified to initiate the model.')
+
+            nest_coef_variation_dict, nest_num_param_dict = parse_formula(nest_formula, dataset.datasets['nest'])
+            item_coef_variation_dict, item_num_param_dict = parse_formula(item_formula, dataset.datasets['item'])
+
+        else:
+            # check for conflicting information.
+            if (nest_formula is not None) or (item_formula is not None):
+                raise ValueError('You should not specify {item, nest}_formula and {item, nest}_coef_variation_dict at the same time.')
+            # make sure that the research specifies all the required information.
+            if (nest_coef_variation_dict is None) or (item_coef_variation_dict is None):
+                raise ValueError('You should specify the {item, nest}_coef_variation_dict to initiate the model.')
+            if (nest_num_param_dict is None) or (item_num_param_dict is None):
+                raise ValueError('You should specify the {item, nest}_num_param_dict to initiate the model.')
+
         super(NestedLogitModel, self).__init__()
-        self.category_to_item = category_to_item
-        self.category_coef_variation_dict = category_coef_variation_dict
-        self.category_num_param_dict = category_num_param_dict
+        self.nest_to_item = nest_to_item
+        self.nest_coef_variation_dict = nest_coef_variation_dict
+        self.nest_num_param_dict = nest_num_param_dict
         self.item_coef_variation_dict = item_coef_variation_dict
         self.item_num_param_dict = item_num_param_dict
         self.num_users = num_users
 
-        self.categories = list(category_to_item.keys())
-        self.num_categories = len(self.categories)
-        self.num_items = sum(len(items) for items in category_to_item.values())
+        self.nests = list(nest_to_item.keys())
+        self.num_nests = len(self.nests)
+        self.num_items = sum(len(items) for items in nest_to_item.values())
 
-        # category coefficients.
-        self.category_coef_dict = self._build_coef_dict(self.category_coef_variation_dict,
-                                                        self.category_num_param_dict,
-                                                        self.num_categories)
+        # nest coefficients.
+        self.nest_coef_dict = self._build_coef_dict(self.nest_coef_variation_dict,
+                                                    self.nest_num_param_dict,
+                                                    self.num_nests)
 
         # item coefficients.
         self.item_coef_dict = self._build_coef_dict(self.item_coef_variation_dict,
@@ -93,7 +136,7 @@ class NestedLogitModel(nn.Module):
         if self.shared_lambda:
             self.lambda_weight = nn.Parameter(torch.ones(1), requires_grad=True)
         else:
-            self.lambda_weight = nn.Parameter(torch.ones(self.num_categories) / 2, requires_grad=True)
+            self.lambda_weight = nn.Parameter(torch.ones(self.num_nests) / 2, requires_grad=True)
         # breakpoint()
         # self.iv_weights = nn.Parameter(torch.ones(1), requires_grad=True)
         # used to warn users if forgot to call clamp.
@@ -124,7 +167,7 @@ class NestedLogitModel(nn.Module):
                          num_items: int) -> nn.ModuleDict:
         """Builds a coefficient dictionary containing all trainable components of the model, mapping coefficient names
             to the corresponding Coefficient Module.
-            num_items could be the actual number of items or the number of categories depends on the use case.
+            num_items could be the actual number of items or the number of nests depends on the use case.
             NOTE: torch-choice users don't directly interact with this method.
 
         Args:
@@ -133,8 +176,7 @@ class NestedLogitModel(nn.Module):
             num_param_dict (Dict[str, int]): a dictionary mapping coefficient names to the number of parameters in this
                 coefficient. Be aware that, for example, if there is one K-dimensional coefficient for every user, then
                 the `num_param` should be K instead of K x number of users.
-            num_items (int): the total number of items in the prediction problem. `num_items` should be the number of
-                categories if _build_coef_dict() is used for category-level prediction.
+            num_items (int): the total number of items in the prediction problem. `num_items` should be the number of nests if _build_coef_dict() is used for nest-level prediction.
 
         Returns:
             nn.ModuleDict: a PyTorch ModuleDict object mapping from coefficient names to training Coefficient.
@@ -148,20 +190,6 @@ class NestedLogitModel(nn.Module):
                                               num_params=num_params)
         return nn.ModuleDict(coef_dict)
 
-    # def _check_input_shapes(self, category_x_dict, item_x_dict, user_index, item_availability) -> None:
-    #     T = list(category_x_dict.values())[0].shape[0]  # batch size.
-    #     for var_type, x_category in category_x_dict.items():
-    #         x_item = item_x_dict[var_type]
-    #         assert len(x_item.shape) == len(x_item.shape) == 3
-    #         assert x_category.shape[0] == x_item.shape[0]
-    #         assert x_category.shape == (T, self.num_categories, self.category_num_param_dict[var_type])
-    #         assert x_item.shape == (T, self.num_items, self.item_num_param_dict[var_type])
-
-    #     if (user_index is not None) and (self.num_users is not None):
-    #         assert user_index.shape == (T,)
-
-    #     if item_availability is not None:
-    #         assert item_availability.shape == (T, self.num_items)
 
     def forward(self, batch: ChoiceDataset) -> torch.Tensor:
         """An standard forward method for the model, the user feeds a ChoiceDataset batch and the model returns the
@@ -178,13 +206,13 @@ class NestedLogitModel(nn.Module):
             torch.Tensor: a tensor of shape (num_trips, num_items) including the log probability
             of choosing item i in trip t.
         """
-        return self._forward(batch['category'].x_dict,
+        return self._forward(batch['nest'].x_dict,
                              batch['item'].x_dict,
                              batch['item'].user_index,
                              batch['item'].item_availability)
 
     def _forward(self,
-                 category_x_dict: Dict[str, torch.Tensor],
+                 nest_x_dict: Dict[str, torch.Tensor],
                  item_x_dict: Dict[str, torch.Tensor],
                  user_index: Optional[torch.LongTensor] = None,
                  item_availability: Optional[torch.BoolTensor] = None
@@ -193,11 +221,15 @@ class NestedLogitModel(nn.Module):
         Let n denote the ID of the user involved in trip t, then P[t, i] = P_{ni} on page 86 of the
         book "discrete choice methods with simulation" by Train.
 
+        The `_forward` method is an internal API, users should refer to the `forward` method.
+
         Args:
-            x_category (torch.Tensor): a tensor with shape (num_trips, num_categories, *) including
-                features of all categories in each trip.
-            x_item (torch.Tensor): a tensor with shape (num_trips, num_items, *) including features
-                of all items in each trip.
+            nest_x_dict (torch.Tensor): a dictionary mapping from nest-level feature names to the corresponding feature tensor.
+
+            item_x_dict (torch.Tensor): a dictionary mapping from item-level feature names to the corresponding feature tensor.
+
+                More details on the shape of the tensors can be found in the docstring of the `x_dict` method of `ChoiceDataset`.
+
             user_index (torch.LongTensor): a tensor of shape (num_trips,) indicating which user is
                 making decision in each trip. Setting user_index = None assumes the same user is
                 making decisions in all trips.
@@ -212,24 +244,24 @@ class NestedLogitModel(nn.Module):
             of choosing item i in trip t.
         """
         if self.shared_lambda:
-            self.lambdas = self.lambda_weight.expand(self.num_categories)
+            self.lambdas = self.lambda_weight.expand(self.num_nests)
         else:
             self.lambdas = self.lambda_weight
 
         # if not self._clamp_called_flag:
         #     warnings.warn('Did you forget to call clamp_lambdas() after optimizer.step()?')
 
-        # The overall utility of item can be decomposed into V[item] = W[category] + Y[item] + eps.
+        # The overall utility of item can be decomposed into V[item] = W[nest] + Y[item] + eps.
         T = list(item_x_dict.values())[0].shape[0]
         device = list(item_x_dict.values())[0].device
-        # compute category-specific utility with shape (T, num_categories).
-        W = torch.zeros(T, self.num_categories).to(device)
+        # compute nest-specific utility with shape (T, num_nests).
+        W = torch.zeros(T, self.num_nests).to(device)
 
-        if 'intercept' in self.category_coef_variation_dict.keys():
-            category_x_dict['intercept'] = torch.ones((T, self.num_categories, 1)).to(device)
+        if 'intercept' in self.nest_coef_variation_dict.keys():
+            nest_x_dict['intercept'] = torch.ones((T, self.num_nests, 1)).to(device)
 
-        for var_type, coef in self.category_coef_dict.items():
-            W += coef(category_x_dict[var_type], user_index)
+        for var_type, coef in self.nest_coef_dict.items():
+            W += coef(nest_x_dict[var_type], user_index)
 
         # compute item-specific utility (T, num_items).
         Y = torch.zeros(T, self.num_items).to(device)
@@ -240,38 +272,38 @@ class NestedLogitModel(nn.Module):
             Y[~item_availability] =torch.finfo(Y.dtype).min / 2
 
         # =============================================================================
-        # compute the inclusive value of each category.
+        # compute the inclusive value of each nest.
         inclusive_value = dict()
-        for k, Bk in self.category_to_item.items():
+        for k, Bk in self.nest_to_item.items():
             # for nest k, divide the Y of all items in Bk by lambda_k.
             Y[:, Bk] /= self.lambdas[k]
-            # compute inclusive value for category k.
+            # compute inclusive value for nest k.
             # mask out unavilable items.
             inclusive_value[k] = torch.logsumexp(Y[:, Bk], dim=1, keepdim=False)  # (T,)
-        # boardcast inclusive value from (T, num_categories) to (T, num_items).
-        # for trip t, I[t, i] is the inclusive value of the category item i belongs to.
+        # boardcast inclusive value from (T, num_nests) to (T, num_items).
+        # for trip t, I[t, i] is the inclusive value of the nest item i belongs to.
         I = torch.zeros(T, self.num_items).to(device)
-        for k, Bk in self.category_to_item.items():
+        for k, Bk in self.nest_to_item.items():
             I[:, Bk] = inclusive_value[k].view(-1, 1)  # (T, |Bk|)
 
-        # logP_item[t, i] = log P(ni|Bk), where Bk is the category item i is in, n is the user in trip t.
+        # logP_item[t, i] = log P(ni|Bk), where Bk is the nest item i is in, n is the user in trip t.
         logP_item = Y - I  # (T, num_items)
 
         # =============================================================================
-        # logP_category[t, i] = log P(Bk), for item i in trip t, the probability of choosing the nest/bucket
-        # item i belongs to. logP_category has shape (T, num_items)
+        # logP_nest[t, i] = log P(Bk), for item i in trip t, the probability of choosing the nest/bucket
+        # item i belongs to. logP_nest has shape (T, num_items)
         # logit[t, i] = W[n, k] + lambda[k] I[n, k], where n is the user involved in trip t, k is
-        # the category item i belongs to.
+        # the nest item i belongs to.
         logit = torch.zeros(T, self.num_items).to(device)
-        for k, Bk in self.category_to_item.items():
+        for k, Bk in self.nest_to_item.items():
             logit[:, Bk] = (W[:, k] + self.lambdas[k] * inclusive_value[k]).view(-1, 1)  # (T, |Bk|)
-        # only count each category once in the logsumexp within the category level model.
-        cols = [x[0] for x in self.category_to_item.values()]
-        logP_category = logit - torch.logsumexp(logit[:, cols], dim=1, keepdim=True)
+        # only count each nest once in the logsumexp within the nest level model.
+        cols = [x[0] for x in self.nest_to_item.values()]
+        logP_nest = logit - torch.logsumexp(logit[:, cols], dim=1, keepdim=True)
 
         # =============================================================================
         # compute the joint log P_{ni} as in the textbook.
-        logP = logP_item + logP_category
+        logP = logP_item + logP_nest
         self._clamp_called_flag = False
         return logP
 
